package xyz.zhouxy.plusone.validator.validator2;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;

import lombok.AccessLevel;
import lombok.AllArgsConstructor;
import xyz.zhouxy.plusone.constant.RegexConsts;
import xyz.zhouxy.plusone.exception.InvalidInputException;
import xyz.zhouxy.plusone.util.RegexUtil;

public class ValueValidator<DTO, PROPERTY> {
    Function<DTO, PROPERTY> getter;
    List<RuleInfo<PROPERTY, ?>> rules = new ArrayList<>();

    public ValueValidator(Function<DTO, PROPERTY> getter) {
        this.getter = getter;
    }

    private void withRule(Predicate<PROPERTY> condition, String errMsg) {
        withRule(condition, value -> new InvalidInputException(errMsg));
    }

    private <E extends RuntimeException> void withRule(Predicate<PROPERTY> condition,
            Function<PROPERTY, E> exceptionCreator) {
        this.rules.add(new RuleInfo<>(condition, exceptionCreator));
    }

    // ====================
    // ====== Object ======
    // ====================

    // ====== notNull =====

    public ValueValidator<DTO, PROPERTY> notNull(String errMsg) {
        return notNull(value -> new InvalidInputException(errMsg));
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> notNull(Supplier<E> exceptionCreator) {
        return notNull(value -> exceptionCreator.get());
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> notNull(Function<PROPERTY, E> exceptionCreator) {
        withRule(Objects::nonNull, exceptionCreator);
        return this;
    }

    // ====== isNull =====

    public ValueValidator<DTO, PROPERTY> isNull(String errMsg) {
        return isNull(value -> new InvalidInputException(errMsg));
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> isNull(Supplier<E> exceptionCreator) {
        return isNull(value -> exceptionCreator.get());
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> isNull(Function<PROPERTY, E> exceptionCreator) {
        withRule(Objects::isNull, exceptionCreator);
        return this;
    }

    // ===== equals =====

    public ValueValidator<DTO, PROPERTY> equalsThat(Object that) {
        return equalsThat(that, value -> new InvalidInputException(String.format("(%s) 必须与 (%s) 相等", value, that)));
    }

    public ValueValidator<DTO, PROPERTY> equalsThat(Object that, String errMsg) {
        return equalsThat(that, value -> new InvalidInputException(errMsg));
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> equalsThat(
            Object that, Supplier<E> exceptionCreator) {
        return equalsThat(that, value -> exceptionCreator.get());
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> equalsThat(
            Object that, Function<PROPERTY, E> exceptionCreator) {
        withRule(value -> Objects.equals(value, that), exceptionCreator);
        return this;
    }

    // ===== state =====

    public ValueValidator<DTO, PROPERTY> state(Predicate<PROPERTY> condition) {
        return state(condition, value -> new InvalidInputException("无效的用户输入"));
    }

    public ValueValidator<DTO, PROPERTY> state(Predicate<PROPERTY> condition, String errMsg) {
        return state(condition, value -> new InvalidInputException(errMsg));
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> state(
            Predicate<PROPERTY> condition,
            Supplier<E> exceptionCreator) {
        return state(condition, value -> exceptionCreator.get());
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> state(
            Predicate<PROPERTY> condition,
            Function<PROPERTY, E> exceptionCreator) {
        withRule(condition, exceptionCreator);
        return this;
    }

    // =================
    // ====== int ======
    // =================

    public ValueValidator<DTO, PROPERTY> between(int min, int max) {
        return between(min, max, String.format("数值不在 %d 和 %d 之间", min, max));
    }

    public ValueValidator<DTO, PROPERTY> between(int min, int max, String errMsg) {
        return between(min, max, value -> new InvalidInputException(errMsg));
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> between(int min, int max,
            Supplier<E> exceptionCreator) {
        return between(min, max, value -> exceptionCreator.get());
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> between(int min, int max,
            Function<PROPERTY, E> exceptionCreator) {
        withRule(value -> ((int) value >= min && (int) value < max), exceptionCreator);
        return this;
    }

    // ================================
    // ====== Collection, String ======
    // ================================

    // ====== notEmpty =====

    public ValueValidator<DTO, PROPERTY> notEmpty(String errMsg) {
        return notEmpty(value -> new InvalidInputException(errMsg));
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> notEmpty(Supplier<E> exceptionCreator) {
        return notEmpty(value -> exceptionCreator.get());
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> notEmpty(Function<PROPERTY, E> exceptionCreator) {
        withRule(value -> {
            if (value == null) {
                return false;
            }
            if (value instanceof Collection) {
                return !((Collection<?>) value).isEmpty();
            }
            if (value instanceof String) {
                return !((String) value).isEmpty();
            }
            return false;
        }, exceptionCreator);
        return this;
    }

    // ====== isEmpty =====

    public ValueValidator<DTO, PROPERTY> isEmpty(String errMsg) {
        return isEmpty(value -> new InvalidInputException(errMsg));
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> isEmpty(Supplier<E> exceptionCreator) {
        return isEmpty(value -> exceptionCreator.get());
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> isEmpty(Function<PROPERTY, E> exceptionCreator) {
        withRule(value -> {
            if (value == null) {
                return false;
            }
            if (value instanceof Collection) {
                return ((Collection<?>) value).isEmpty();
            }
            if (value instanceof String) {
                return ((String) value).isEmpty();
            }
            return false;
        }, exceptionCreator);
        return this;
    }

    // =====================
    // ====== boolean ======
    // =====================

    // ====== isTrue ======

    public ValueValidator<DTO, PROPERTY> isTrue() {
        return isTrue(value -> new InvalidInputException("Value must be true."));
    }

    public ValueValidator<DTO, PROPERTY> isTrue(String errMsg) {
        return isTrue(value -> new InvalidInputException(errMsg));
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> isTrue(Supplier<E> exceptionCreator) {
        return isTrue(value -> exceptionCreator.get());
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> isTrue(Function<PROPERTY, E> exceptionCreator) {
        withRule(Boolean.TRUE::equals, exceptionCreator);
        return this;
    }

    // ====== isFalse ======

    public ValueValidator<DTO, PROPERTY> isFalse() {
        return isFalse(value -> new InvalidInputException("Value must be true."));
    }

    public ValueValidator<DTO, PROPERTY> isFalse(String errMsg) {
        return isFalse(value -> new InvalidInputException(errMsg));
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> isFalse(Supplier<E> exceptionCreator) {
        return isFalse(value -> exceptionCreator.get());
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> isFalse(Function<PROPERTY, E> exceptionCreator) {
        withRule(Boolean.FALSE::equals, exceptionCreator);
        return this;
    }

    // ====================
    // ====== String ======
    // ====================

    // ===== matches =====

    public ValueValidator<DTO, PROPERTY> matches(String regex, String errMsg) {
        return matches(regex, value -> new InvalidInputException(errMsg));
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> matches(String regex,
            Supplier<E> exceptionCreator) {
        return matches(regex, value -> exceptionCreator.get());
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> matches(String regex,
            Function<PROPERTY, E> exceptionCreator) {
        withRule(input -> RegexUtil.matches((String) input, regex), exceptionCreator);
        return this;
    }

    // ===== matchesOr =====

    public ValueValidator<DTO, PROPERTY> matchesOr(String[] regexs, String errMsg) {
        return matchesOr(regexs, value -> new InvalidInputException(errMsg));
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> matchesOr(String[] regexs,
            Supplier<E> exceptionCreator) {
        return matchesOr(regexs, value -> exceptionCreator.get());
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> matchesOr(String[] regexs,
            Function<PROPERTY, E> exceptionCreator) {
        withRule(input -> RegexUtil.matchesOr((String) input, regexs), exceptionCreator);
        return this;
    }

    // ===== matchesOr =====

    public ValueValidator<DTO, PROPERTY> matchesAnd(String[] regexs, String errMsg) {
        withRule(input -> RegexUtil.matchesAnd((String) input, regexs), errMsg);
        return this;
    }

    // ===== email =====

    public ValueValidator<DTO, PROPERTY> email(String errMsg) {
        return email(value -> new InvalidInputException(errMsg));
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> email(Supplier<E> exceptionCreator) {
        return email(value -> exceptionCreator.get());
    }

    public <E extends RuntimeException> ValueValidator<DTO, PROPERTY> email(Function<PROPERTY, E> exceptionCreator) {
        return matches(RegexConsts.EMAIL, exceptionCreator);
    }

    // ========================================================================

    void validateProperty(DTO obj) {
        PROPERTY value = this.getter.apply(obj);
        this.rules.forEach(rule -> rule.validate(value));
    }

    @AllArgsConstructor(access = AccessLevel.PRIVATE)
    private static final class RuleInfo<V, E extends RuntimeException> {
        private final Predicate<V> condition;
        private final Function<V, E> exceptionCreator;

        private void validate(V obj) {
            if (!this.condition.test(obj)) {
                throw exceptionCreator.apply(obj);
            }
        }
    }
}
